! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                   Test for MPI_Issend():  All others Send TO Root
!                        and MPI_Irecv():   Any source/Any tag
!
!
! Issend is the synchronous call to the non-blocking send.
! Section 3.4 of the MPI Standard states that a send that uses synchronous mode
! can be started whether or not a matching receive was posted. The send wil
! only complete when the matching receive has been posted, and has started to
! receive the message.  For tese blocking sends and receives, the communication
! does not complete at either end until both operations rendezvous at the
! communication.  Thus, the Issend is non-local, and no additional
! synchronization is required.
! 
! For an INTRA-communicator, the program selects each node, in turn,
! of the communicator to be the root.  the ROOT receives from all of the
! other nodes (it does NOT receive from itself).
! 
! For an INTER-communicator, there are two groups of nodes; each group being
! termed a sub-communicator.  The nodes in each sub-communicator are numbered
! 0 to (n-1) for an n-node sub-communicator.  So,the MPITEST_current_rank will
! return duplicate node numbers for nodes (0 to k-1) where k is the number of
! nodes in the smaller communicator.
! 
! The program cycles through the nodes in one sub-communicator, sending from
! each selected node to  all of the nodes in the other sub-communicator.  Then
! the program reverses the send and receive role of the two sub-communicators,
! and repeats the process.
! 
! This test initializes the send buffer with the root's rank in the
! communicator (or an appropriate value for the non-integer types.)
! Once the receiving nodes have completed their message receive, they check to
! make sure the current root's rank is in the received buffer.
! 
! This test may be run in any communicator, with any data type, and with
! any non-negative message length.
! 
! The MPITEST environment provides looping over communicator size and
! type, message length, and data type.  The properties of the loops are
! encoded in configuration arrays in the file config.h .  See the
! MPITEST README for further details.
! 
! **********************************************************************
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM TEST
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      INTEGER    TYPE_COUNT
!                               number of data types to test
      INTEGER    TEST_TYPE
!                               index of current data type
      INTEGER    LENGTH_COUNT
!                               number of data lengths to test
      INTEGER    MAX_LENGTH
!                               max message length to test
      INTEGER    MAX_BYTE_LENGTH
!                               max byte length to test
      INTEGER    MAX_BUFF_LENGTH
!                               max length that will fit in buffers
      INTEGER    BYTE_LENGTH
!                               current byte length to test
      INTEGER    LENGTH
!                               current length to test 

      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    LOOP_FAIL
!                               counts number of failures in 1 loop
      INTEGER    ERROR, ERROR2
!                               number of data errors found in buffer
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ROOT
!                               the current root
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    COUNT_LEN
!                               Length of received message from MPI_GET_COUNT
      INTEGER    GRP_LUP
!                               For INTER, index of comm group
      LOGICAL    INTER_FLAG
!                               Is comm under test an INTER communicator
      INTEGER    RECV_SIZE
      INTEGER    LEFT_RECV_SIZE
      INTEGER    RIGHT_RECV_SIZE
!                              For INTER, number of receivers on each side
      INTEGER    SEND_SIZE
      INTEGER    LEFT_SEND_SIZE
      INTEGER    RIGHT_SEND_SIZE
!                              For INTER, number of senders on each side
      INTEGER    SEND_GROUP
      INTEGER    RECV_GROUP
!                              For INTER, current group under test
      INTEGER    NTIMES
!                              Number of times to go through loop
!                              INTER=2, INTRA=1
      INTEGER    SENDERS
      INTEGER    RECEIVERS
      INTEGER    GET_FROM
!                              Number of senders/receivers
      LOGICAL    PRINT_NODE
!                              Is node to print informational messages

      INTEGER    I, J, K, L, M

      MPITEST_BUF_TYPE SEND_BUFFER(MAX_BUFF_SIZE+1)
      MPITEST_BUF_TYPE RECV_BUFFER(MAX_BUFF_SIZE+1)
      INTEGER    ERRSIZE

      INTEGER    RECV_STAT(MPI_STATUS_SIZE)
      INTEGER    SEND_STAT(MPI_STATUS_SIZE)

      INTEGER    RECV_REQ
      INTEGER    SEND_REQ

      MPITEST_BUF_TYPE VALUE

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 98   FORMAT(A,INT_FMT,A,INT_FMT)
 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)
 189  FORMAT(INT_FMT,A,A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)

!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Issend_ator'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
      CALL MPITEST_NUM_DATATYPES(TYPE_COUNT)
      CALL MPITEST_NUM_MESSAGE_LENGTHS(LENGTH_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

!
!     Loop through communicators
!
      DO 100 I=1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(I, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(I, COMM_TYPE)

        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)

!       Skip to end of loop if this node is not part of current
!       communicator

        IF (COMM .EQ. MPI_COMM_NULL) GO TO 201

!       Test for INTER-communicator and initialize appropriately if so

        CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          INFOBUF=' '
          WRITE(INFOBUF,99) 'MPI_COMM_TEST_INTER() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          FAIL = FAIL + 1
        END IF

        PRINT_NODE = .FALSE.

        IF (INTER_FLAG) THEN
          IF (MPITEST_INTER .EQ. 0) THEN
!           Left side of intercommunicator
            CALL MPI_COMM_SIZE(COMM, SEND_SIZE, IERR)
            CALL MPI_COMM_REMOTE_SIZE(COMM, RECV_SIZE, IERR)

            LEFT_SEND_SIZE = SEND_SIZE
            LEFT_RECV_SIZE = RECV_SIZE
            RIGHT_SEND_SIZE = RECV_SIZE
            RIGHT_RECV_SIZE = SEND_SIZE

            IF (MPITEST_CURRENT_RANK .EQ. 0) PRINT_NODE = .TRUE.

          END IF

          IF (MPITEST_INTER .EQ. 1) THEN
!           Right side of intercommunicator
            CALL MPI_COMM_SIZE(COMM, SEND_SIZE, IERR)
            CALL MPI_COMM_REMOTE_SIZE(COMM, RECV_SIZE, IERR)

            LEFT_SEND_SIZE = RECV_SIZE
            LEFT_RECV_SIZE = SEND_SIZE
            RIGHT_SEND_SIZE = SEND_SIZE
            RIGHT_RECV_SIZE = RECV_SIZE

          END IF

        ELSE
!       Intracommunicator
          IF (MPITEST_CURRENT_RANK .EQ. 0) PRINT_NODE = .TRUE.
        END IF

!
!       Loop through data types.
!
        DO 200 J=1,TYPE_COUNT

          CALL MPITEST_GET_DATATYPE(J, TEST_TYPE)
!
!         Convert the number of bytes in the maximum length message
!         into the number of elements of the current type 
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, MAX_BYTE_LENGTH, &
                MAX_LENGTH)
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                 MAX_BUFF_SIZE*MPITEST_BUF_EXTENT, MAX_BUFF_LENGTH)

          IF (MAX_LENGTH .GE. (MAX_BUFF_LENGTH - 1)) &
               MAX_LENGTH = MAX_BUFF_LENGTH - 1


!
!         Loop through data lengths
!
          DO 300 K=1,LENGTH_COUNT
            CALL MPITEST_GET_MESSAGE_LENGTH(K, BYTE_LENGTH)
            CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, BYTE_LENGTH, &
                LENGTH)
            IF (LENGTH .GT. MAX_LENGTH) LENGTH = MAX_LENGTH

!
!           Loop over different processes being the root
!
            IF (INTER_FLAG) THEN
              NTIMES = 2
            ELSE
              NTIMES = 1
            END IF

            DO 350 GRP_LUP=1, NTIMES
!             If intercommunicators, test each side
              SEND_GROUP = GRP_LUP - 1
              RECV_GROUP = 2 - GRP_LUP

              RECEIVERS = COMM_SIZE
              IF (INTER_FLAG .AND. (GRP_LUP .EQ. 1)) THEN
                RECEIVERS = LEFT_RECV_SIZE
                SENDERS = LEFT_SEND_SIZE
              ELSE IF (INTER_FLAG .AND. (GRP_LUP .EQ. 2)) THEN
                RECEIVERS = RIGHT_RECV_SIZE
                SENDERS = RIGHT_SEND_SIZE
              END IF

              DO 400 ROOT=0,RECEIVERS-1
!               Print an informational message
                IF (PRINT_NODE) THEN
                  INFOBUF=' '
                  WRITE (INFOBUF,179)  &
                    'commindex', COMM_INDEX, ', commsize', &
                    COMM_SIZE, ', commtype', COMM_TYPE, ', root', ROOT, &
                    ', datatype', TEST_TYPE, ', length',  LENGTH
                  CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
                END IF

!
!               The root node receives a message for each send node.
!
                IF (((INTER_FLAG .EQV. .FALSE.) .AND.  &
                     (MPITEST_CURRENT_RANK .EQ. ROOT)) .OR. &
                    ((INTER_FLAG .EQV. .TRUE.) .AND. &
                     (MPITEST_INTER .EQ. RECV_GROUP) .AND. &
                     (MPITEST_CURRENT_RANK .EQ. ROOT))) THEN

                  GET_FROM = COMM_SIZE - 1
                  IF (INTER_FLAG) GET_FROM = SENDERS

                  DO 500 M=0, GET_FROM - 1
                    LOOP_CNT = LOOP_CNT + 1

!
!                 Initialize expected receive data
!
                    CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, &
                      TEST_TYPE)
                    CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH+1, VALUE, &
                      RECV_BUFFER)
                    LOOP_CNT = LOOP_CNT + 1
                    LOOP_FAIL = 0


                    CALL MPI_IRECV(RECV_BUFFER, LENGTH+1, &
                      MPITEST_MPI_DATATYPES(TEST_TYPE), MPI_ANY_SOURCE, &
                      MPI_ANY_TAG, COMM, RECV_REQ, IERR)
                    IF (IERR .NE. MPI_SUCCESS) THEN
                      INFOBUF=' '
                      WRITE(INFOBUF,99) 'MPI_RECV() returned', IERR
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      INFOBUF=' '
                      CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                           ERR)
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL,ERRORSTRING)
                      LOOP_FAIL = LOOP_FAIL + 1
                    END IF

                    CALL MPI_WAIT(RECV_REQ, RECV_STAT, IERR)
                    IF (IERR .NE. MPI_SUCCESS) THEN
                      INFOBUF=' '
                      WRITE(INFOBUF,99) 'MPI_WAIT(IRECV) returned', IERR
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      INFOBUF=' '
                      CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                           ERR)
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL,ERRORSTRING)
                      IF (RECV_STAT(MPI_ERROR) .NE. MPI_SUCCESS) THEN
                        INFOBUF=' '
                        CALL MPI_ERROR_STRING(RECV_STAT(MPI_ERROR), &
                           INFOBUF, ERRSIZE, ERR)
                        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      END IF
                      FAIL = FAIL + 1
                    END IF
!
!                   Check data for errors and overflow
!
                    CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                      RECV_STAT(MPI_SOURCE), TEST_TYPE)
                    CALL MPITEST_BUFFER_ERRORS(TEST_TYPE, LENGTH, VALUE, &
                      RECV_BUFFER, ERROR)
                    CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, &
                      TEST_TYPE)
                    CALL MPITEST_BUFFER_ERRORS_OV(TEST_TYPE, LENGTH, &
                      VALUE, RECV_BUFFER, ERROR2)
                    IF ((ERROR + ERROR2) .NE. 0) THEN
                      INFOBUF=' '
                      WRITE (INFOBUF,189) ERROR+ERROR2, &
                        ' errors in buffer, ', &
                        'commindex',COMM_INDEX,', commsize', COMM_SIZE, &
                        ', commtype', COMM_TYPE, ', root', ROOT, &
                        ', datatype', TEST_TYPE, ', length',  LENGTH
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      LOOP_FAIL = LOOP_FAIL + 1
                    END IF

!
!                   Call MPI_GET_COUNT and compare value with length received
!
                    CALL MPI_GET_COUNT(RECV_STAT,  &
                      MPITEST_MPI_DATATYPES(TEST_TYPE), COUNT_LEN, IERR)
                    IF (IERR .NE. MPI_SUCCESS) THEN
                      INFOBUF=' '
                      WRITE(INFOBUF,99) 'MPI_GET_COUNT() returned', IERR
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      INFOBUF=' '
                      CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE,  &
                           ERR)
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL,ERRORSTRING)
                      LOOP_FAIL = LOOP_FAIL + 1
                    END IF

                    IF (LENGTH - COUNT_LEN .NE. 0) THEN
                      INFOBUF=' '
                      WRITE(INFOBUF,98) 'Unexpected receive length', &
                          COUNT_LEN, ', expected', LENGTH
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      INFOBUF=' '
                      LOOP_FAIL = LOOP_FAIL + 1
                    END IF

!
!                   Check tag values.
!
                    IF (RECV_STAT(MPI_TAG) .NE.  &
                        RECV_STAT(MPI_SOURCE)+MPITEST_CURRENT_RANK) THEN
                      INFOBUF=' '
                      WRITE(INFOBUF,98) 'Unexpected tag value', &
                        RECV_STAT(MPI_TAG), ', expected',  &
                        RECV_STAT(MPI_SOURCE)+MPITEST_CURRENT_RANK
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      INFOBUF=' '
                      LOOP_FAIL = LOOP_FAIL + 1
                    END IF

                    IF (LOOP_FAIL .NE. 0) FAIL = FAIL + 1

 500              CONTINUE
!                        Receive from all senders

                END IF
!                    Receiver

!
!               Others send the message.
!
                IF (((INTER_FLAG .EQV. .FALSE.) .AND.  &
                     (MPITEST_CURRENT_RANK .NE. ROOT)) .OR. &
                    ((INTER_FLAG .EQV. .TRUE.) .AND. &
                     (MPITEST_INTER .EQ. SEND_GROUP))) THEN
                  CALL MPITEST_DATATEMPLATE_INIT(VALUE,  &
                    MPITEST_CURRENT_RANK, TEST_TYPE)
                  CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH+1, &
                    VALUE, SEND_BUFFER)

                  CALL MPI_ISSEND(SEND_BUFFER, LENGTH, &
                    MPITEST_MPI_DATATYPES(TEST_TYPE), ROOT,  &
                    MPITEST_CURRENT_RANK + ROOT, COMM, SEND_REQ, IERR)
                  IF (IERR .NE. MPI_SUCCESS) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,99) 'MPI_ISSEND() returned', IERR
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                         ERR)
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                    FAIL = FAIL + 1
                  END IF

                  CALL MPI_WAIT(SEND_REQ, SEND_STAT, IERR)
                  IF (IERR .NE. MPI_SUCCESS) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,99) 'MPI_WAIT(ISSEND) returned', IERR
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                         ERR)
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                    IF (SEND_STAT(MPI_ERROR) .NE. MPI_SUCCESS) THEN
                      INFOBUF=' '
                      CALL MPI_ERROR_STRING(SEND_STAT(MPI_ERROR), &
                             INFOBUF, ERRSIZE, ERR)
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    END IF
                    FAIL = FAIL + 1
                  END IF

                END IF
!                  Sender

 400          CONTINUE
!                   End root loop


 350        CONTINUE
!                   End INTER send group loop
      
 300      CONTINUE
!                 End data length loop
      
 200    CONTINUE
!               End data type loop

 201    CONTINUE

        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)
        CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)

 100  CONTINUE
!             End communicator loop

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END
