! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                    Error test for MPI_SENDRECV()
!
! This test verifies that the correct error is returned if MPI_SENDRECV is
! called with an invalid argument.
!
! 1)  Destination non-existent rank (rank negative)
! 2)  Destination rank greater than group size
! 3)  Call with MPI_COMM_NULL
! 4)  Call with a freed communicator
! 5)  Use a tag < 0
! 6)  Use a tag larger than MPI_TAG_UB
! 7)  Use a negative count argument
! 8)  Use a non-existent data type
! 9)  Use mismatching type signature at sender and receiver
!
! In all cases, expect to receive appropriate error.
!
! Test history:
!  6a  06/18/96     JH  Created
!
!-----------------------------------------------------------------------
!  NOTES:
!
!       For the Sendrecv tests there are two versions of this test.
!       Each error condition is tested seperately on the send and receive side
!       of the call.  The "a" version of the test sets the error condition
!       on the send side of the call.  The "b" version of the test sets the 
!       error condition on the receive side of the test.
! **********************************************************************
#include "f90options.h"

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER PASS
      INTEGER FAIL              
      INTEGER VERIFY
!                           return value from MPI calls
      INTEGER IERR
      INTEGER ERR
!                           error class of IERR
      INTEGER ERRORCLASS
!                           used to split a communicator
      INTEGER COLOR
      INTEGER SIZE 

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
      CHARACTER*32   TESTNAME


!                           MPI communicators
      INTEGER COMM

!                           Input/Output to/from  SENDRECV
      INTEGER INBUFF(10)
      INTEGER OUTBUFF(10)

      INTEGER STATUS(MPI_STATUS_SIZE)
      INTEGER TAG
      LOGICAL FLAG

 98   FORMAT(A,INT_FMT,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)

!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      TESTNAME='MPI_SENDRECV_err6a'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0
      VERIFY = 0
! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 99) ' MPI_ERRORHANDLER_SET returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Invalid tag (> MPI_TAG_UB)
!
      IERR = 0
      CALL MPI_ATTR_GET(MPI_COMM_WORLD, MPI_TAG_UB, TAG, FLAG, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_ATTR_GET() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      IF (FLAG .NEQV. .TRUE.) THEN
        INFOBUF = 'MPI_ATTR_GET() FLAG returned .FALSE.'
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!---------------------------------   SENDRECV  ------------------------------
!
!  Invalid tag (> MPI_TAG_UB)
!
      IERR = 0

      CALL MPI_SENDRECV(INBUFF,  &
                        1,  &
                        MPI_INTEGER,  &
                        MPITEST_NUMP-1,  &
                        TAG + 1,  &
                        OUTBUFF, &
                        1, &
                        MPI_INTEGER, &
                        MPI_PROC_NULL, &
                        0, &
                        MPI_COMM_WORLD, &
                        STATUS, &
                        IERR)

      IF (IERR .EQ. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 98) 'MPI_SENDRECV() w/ SEND TAG (', TAG,  &
        ')   > MPI_TAG_UB   did not FAIL'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      ELSE
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
        IF (ERRORCLASS .NE. MPI_ERR_TAG) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF, 98) &
            'MPI_SENDRECV() w/ SEND  TAG > MPI_TAG_UB returned',   &
            ERRORCLASS, ' expected MPI_ERR_TAG'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ELSE
          PASS = PASS + 1
        END IF
      END IF
!
!  Report overall results
!
      CALL MPITEST_REPORT(PASS, FAIL, VERIFY, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
